/** ----------------------------------- -o-
 * OSCReceive.java  (mxj object)
 *
 *   Receive and decapsulate OSC commands intended for a particular
 *     client ID at a particular port.
 *   Modeled after Sciss server example.
 *
 *
 * OBJECT ARGUMENTS: <port> <clientID> [<oscdomain>]
 *
 * INCOMING OSC FORMAT: 
 *   /<oscdomain>/C/<clientID>/<channel> [<command_token> [<args(s)>]]
 *
 * RECEIVED MAX MESSAGE FORMAT: <channel> [<command_token> [<arg(s)>]]
 *
 *
 * LIBRARY DEPENDENCIES: 
 *   NetUtil -- http://www.sciss.de/netutil/
 *   max.jar -- /Applications/Max5/Cycling '74/java/lib/max.jar
 *
 * BUILD AND INSTALL: 
 *   1) javac -classpath <cp> OSCReceive.java
 *        <cp> = colon separated list of paths of jar file library dependencies.
 *   2) Place resulting class files in any directory under the root directory 
 *        containing the Max patch that uses "mxj OSCReceive.java".
 *      
 *
 * MESSAGES--  
 *   open / close
 *   bang
 *   oscdomain
 *
 * METHODS--
 *   setoscdomain
 *   outletIPInfo
 *   parseCommand
 *   OSCargs2string
 *
 *
 * NB  A change of oscdomain will not be recognized by an active
 *     connection.  Close and re-open the connection to register the change.
 *
 * XXX  Output of OSCReceive must be processed by Max object "sprintf %s"
 *      to split output containing whitespace into list of strings that
 *      may be indexed and separated.
 * 
 * ----------------------------------------------------------------------
 * Copyright © 2011-2012, David Reeder.  All rights reserved.
 * Please see LICENSE file in this distribution for license details.
 * Contact: osc@davidreeder.info
 * ----------------------------------------------------------------------
 */ 	 								// -o--

import com.cycling74.max.*;

import de.sciss.net.*;

import java.net.InetAddress;
import java.net.SocketAddress;
import java.io.IOException;



public class OSCReceive extends MaxObject
{ 							
  //static private DebugLabel_M z = new DebugLabel_M("OSCReceive", true);
  //static private ToolBox    	t = new ToolBox();

  int verbose = 1;


  private String 	USAGE = "Usage: OSCReceive <port> <clientID> [<oscdomain>]";
  private String 	version = "0.3";

  private String 	MXJERR = "MXJ argument error: ";


  private OSCServer 	server = null;
  private OSCListener 	listener = null;

  InetAddress 		recv_ipaddr = null;
  private int 		recv_port;

  private String 	oscdomain = "domain"; 	// DEFAULT
  private String 	recv_prefix = null;
  private String 	recv_clientId;




  //------------------------------------------------------------------- -o--
  // constructors
  //

  public OSCReceive(Atom[] args) 
  { 
    //zsm("OSCReceive");
    zsm("");

    zi("version " + version + ".  osc@davidreeder.info");

    if (2 <= verbose)  { zarglist(args); }


    //
    // parse args and initialize
    //
    if ((2 > args.length) || (3 < args.length)) { 
      zerr(MXJERR + "Improper number of arguments.");
      zferr(USAGE); 
    }

    recv_port 	   = args[0].toInt();
    recv_clientId  = args[1].toString();

    if (2 < args.length) { 
      oscdomain	   = args[2].toString();
    }

    setoscdomain(oscdomain);


    outletIPInfo();
    open();



    //
    // config environment
    //

    declareIO(1, 3);

    String[] ia = new String[] { 	
      "0: (messages)  open/close, oscdomain, bang",
    };

    String[] oa = new String[] { 	
      "0: (message) OSC commands",
      "1: (message) host name",
      "2: (message) host IP address",
    };

    setInletAssist(ia);
    setOutletAssist(oa);
    createInfoOutlet(false);


    zms();
  }  // OSCReceive()




  //------------------------------------------------------------------- -o--
  // messages
  //

  // open()
  // 
  // Instantiate listener.
  //
  public void open() 
  {
    zsm("open");


    // server/listener
    //
    try {
      if (null == server) {
        server = OSCServer.newUsing(OSCServer.UDP, recv_port, false);
	server.start();

	zi("client-ID " + recv_clientId 
	  +" @ "+ recv_ipaddr.getHostAddress() +":"+ recv_port 
	    +" ("+ recv_prefix+")");
      }


      if (null == listener) {
	listener = new OSCListener() {
	  public void messageReceived( 	OSCMessage msg, 
					  SocketAddress addr, 
					  long time )
	  {
	    parseCommand(msg, addr, time);
	  }
	};

	server.addOSCListener(listener);
      }


    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }
      
    zms();
  }  // open();



  public void close() 
  {
    zsm("close");

    try {
      if (null != server) {
	server.stop();

	server.removeOSCListener(listener);
	listener = null;

	server.dispose();
	server = null;

	zi("client-ID " + recv_clientId 
	  +" @ "+ recv_ipaddr.getHostAddress() +":"+ recv_port 
	    +" ("+ recv_prefix +")");
      }

    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }

    zms();
  }  // close()



  public void bang()  { outletIPInfo(); }


  public void oscdomain(String prefix)  { setoscdomain(prefix); }




  //------------------------------------------------------------------- -o--
  // helper methods
  //

  private void setoscdomain(String prefix)
  {
    recv_prefix = "/" + prefix + "/C/" + recv_clientId + "/";
  }



  private void outletIPInfo()  						// -o-
  {
    try {
      recv_ipaddr = InetAddress.getLocalHost();

    } catch (Exception e) {
      e.printStackTrace();
      zsm("bang");
      zerr(e.toString());
      zms();
    }

    outlet(1, recv_ipaddr.getHostName());
    outlet(2, recv_ipaddr.getHostAddress());
  }  // outletIPInfo()



  // parseCommand()
  //
  // Check OSC message prefix represents proper prefix + clientID.
  // Strip slashes ('/') and deliver the remainder with arguments
  //   as a Max message.
  //
  // TBD  Handle bundles.
  // 									// -o-
  private void parseCommand(OSCMessage msg, SocketAddress addr, long time)
  {
    zsm("parseCommand");

    String s = "";
    String[] elements;


    if (0 < verbose) { 
      //zi(t.dumpOSC(msg, addr, time)); 
      zi(msg.getName() +" "+ OSCargs2string(msg)
	      +"  <--  "+ recv_clientId 
	      +" @ "+ recv_ipaddr.getHostAddress() +":"+ recv_port); 
    }


    // ignore messages not addressed to this client
    //
    if (! msg.getName().startsWith(recv_prefix))  { return; }

    elements = msg.getName().substring(1).split("/");

    for (int i = 3; i < elements.length; i++)  { s += elements[i] + " "; }

    s += OSCargs2string(msg);


    outlet(0, s);  
      // XXX  Requires "sprintf %s" to separate string array into elements. 

    zms();
  }  // parseCommand()



  //------------------------------ -o-
  private String OSCargs2string(OSCMessage msg) 
  {
    String s = "";
    for (int i = 0; i < msg.getArgCount(); i++)  { s += msg.getArg(i) + " "; }
    return s.trim();
  }



  //------------------------------ -o-
  // local alternative to DebugLabel_M().  z.<method> --> z<method>
  //

  private String  zclass = "OSCReceive",
  		  zfuncname = null;

  private void zsm(String s)  { zfuncname = s; }
  private void zms()  	      { zfuncname = null; }

  private void zpr(String label, String msg, boolean isErr, boolean isFatal)  
  { 
    String s = label +"\t\t\t\t\t"+ zclass +"."+ zfuncname +" :: "+ msg;
    if (isErr)  { error(s); if (isFatal) {bail("");} }
    else        { post(s); }
  }

  private void zd(String msg)  { zpr("DEBUG", msg, false, false); }
  private void zi(String msg)  { zpr("INFO", msg, false, false); }
  private void zerr(String msg)  { zpr("ERR", msg, true, false); }
  private void zferr(String msg)  { zpr("ERR", msg, true, true); }

  private void zarglist(Atom[] a) {
    zi("ARGS(" + a.length + ")\t\t" + Atom.toOneString(a)); 
  }

}  // public class OSCReceive extends MaxObject
  
